import re
import shutil
import pickle
import datetime
import asyncio
import threading

from obj_encrypt import Secret
from apscheduler.schedulers.background import BackgroundScheduler

from . import AioStream
from .route import Route
from ..data import datas
from ..extensions import DisconCyberDBError, WrongFilenameCyberDBError, \
WrongPasswordCyberDBError, BackupCyberDBError, TypeCyberDBError
from ..extensions.signature import Signature


class Server:
    '''
        This class is used to create CyberDB server objects.
    '''

    def __init__(self):
        self._data = {
            'config': {
                'host': None,
                'port': None,
                'password': None
            },
            'db': {}
        }
        self.ips = {'127.0.0.1'}  # ip whitelist
        self.sched = None

    def start(self, host: str = '127.0.0.1', port: int = 9980,
              password: str = None, max_con: int = 500, timeout: int = 0,
              print_log: bool = False, encrypt: bool = False):
        '''
            Starts the CyberDB server in the background, which does not block 
            foreground tasks.
            
            Parameters:
            
                host -- TCP listening host address, such as 127.0.0.1.
                
                port – TCP listening port.
                
                password -- TCP communication password, it is recommended to use 
                a combination of English letters and numbers,
                Up to 32 characters long.
                
                max_con -- the maximum number of concurrency.
                
                timeout -- The timeout for a single connection, in seconds.
                
                print_log -- whether to print the communication log, Fasle does not 
                print it.
                
                encrypt -- Whether to encrypt the communication content, Fasle is 
                not encrypted. The encryption algorithm is AES-256 and the key is 
                password.
                
            Return Type: None
        '''
        t = threading.Thread(target=self.run,
                             args=(host, port, password, max_con, timeout, print_log, encrypt))
        t.daemon = True
        t.start()

    def run(self, host: str = '127.0.0.1', port: int = 9980,
            password: str = None, max_con: int = 500, timeout: int = 0,
            print_log: bool = False, encrypt: bool = False):
        '''
            Running the CyberDB server in the foreground blocks foreground tasks.
            The parameters are the same as the start method.
            
            Return Type: None
        '''
        if not password:
            raise RuntimeError('The password cannot be empty.')

        self._data['config']['host'] = host
        self._data['config']['port'] = port
        self._data['config']['password'] = password
        self._data['config']['max_con'] = max_con
        self._data['config']['timeout'] = timeout
        self._data['config']['print_log'] = print_log
        self._data['config']['encrypt'] = encrypt

        # Responsible for encrypting and decrypting objects.
        secret = Secret(key=password)
        # for digital signature
        signature = Signature(salt=password.encode())
        # Convert TCP data and encrypted objects to each other.
        self._dp = datas.DataParsing(secret, signature, encrypt=encrypt)

        asyncio.run(self._main())

    async def _main(self):
        server = await asyncio.start_server(
            self._listener, self._data['config']['host'],
            self._data['config']['port'],
            limit=2 ** 16,  # 64 KiB
            # the maximum number of queued connections
            backlog=self._data['config']['max_con']
        )

        print('CyberDB server is starting, please wait a few seconds before operation!\n* Started at {}:{}'.format(
            self._data['config']['host'],
            self._data['config']['port']
        ))

        async with server:
            await server.serve_forever()

    async def _listener(self, reader: asyncio.streams.StreamReader,
                        writer: asyncio.streams.StreamWriter):
        '''
            This method is entered as long as a TCP connection is established,
             even if no data is sent.
        '''

        try:
            addr = writer.get_extra_info('peername')
            if self._data['config']['print_log']:
                print('{}  {}:{}  establishes a connection.'.format(
                    datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'),
                    addr[0], addr[1]))

            if self._data['config']['encrypt']:
                # Check if the ip is in the whitelist.
                if self.ips:
                    if addr[0] not in self.ips:
                        if self._data['config']['print_log']:
                            print('{}  The request for {}, the ip is not in the whitelist.'.format(
                                datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'),
                                addr[0]))
                        writer.close()
                        return

            # TCP route of this connection
            stream = AioStream(reader, writer, self._dp)
            route = Route(self._data['db'], self._dp, stream,
                          print_log=self._data['config']['print_log'])

            # If the timeout is set, it will automatically disconnect.
            if self._data['config']['timeout'] == 0:
                await route.find()
            else:
                try:
                    await asyncio.wait_for(route.find(),
                                           timeout=self._data['config']['timeout'])
                except asyncio.TimeoutError:
                    if self._data['config']['print_log']:
                        print('{}  {}:{}  connection timed out.'.format(
                            datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'),
                            addr[0], addr[1]))
                    writer.close()

        except DisconCyberDBError:
            if self._data['config']['print_log']:
                print('{}  {}:{}  Client disconnected.'.format(
                    datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'),
                    addr[0], addr[1])
                )
        except WrongPasswordCyberDBError:
            if self._data['config']['print_log']:
                print('{}  {}:{}  The password entered by the client is incorrect.'.format(
                    datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'),
                    addr[0], addr[1]))

    def set_ip_whitelist(self, ips: list):
        '''
            Set the ip whitelist. When CyberDB encrypts communication, only 
            whitelisted ip connections are allowed.

            This method only works if cyberdb.Server.start(encrypt=True) or 
            cyberdb.Server.run(encrypt=True) are enabled.

            Parameters:

                ips -- the type is a list, the format is ['192.168.1.1', 
                '118.123.89.137']
                
            Return Type: None
        '''
        for ip in ips:
            if not re.match(r'^((2((5[0-5])|([0-4]\d)))|([0-1]?\d{1,2}))(\.((2((5[0-5])|([0-4]\d)))|([0-1]?\d{1,2}))){3}$', ip):
                raise RuntimeError('Please enter a valid ipv4 address.')
        ips = set(ips)
        ips.add('127.0.0.1')
        self.ips = ips

    def set_backup(self, file_name: str = 'data.cdb', cycle: int = 900):
        '''
            Set timed backup. After this operation is set, data persistent backup 
            will be performed at the specified period, and the CyberDB database 
            will be saved to the hard disk.

            parameter:

                file_name -- the name of the file generated by the backup, the file 
                suffix must be .cdb.

                cycle -- The cycle of the cyclic backup, in seconds.
                
            Return Type: None
        '''
        prefix_name, suffix_name = file_name.rsplit('.', 1)

        if suffix_name != 'cdb':
            raise WrongFilenameCyberDBError(
                'Please enter a filename with a .cdb suffix.')

        if type(cycle) != int:
            raise TypeCyberDBError('The type of cycle is not an integer.')

        # If the scheduled task exists, reset it.
        if self.sched:
            self.sched.shutdown(wait=False)

        # The time here is for looping only and does not affect usage anywhere in the world.
        self.sched = BackgroundScheduler(timezone='Asia/Shanghai')
        self.sched.add_job(self.save_db, 'interval',
                           seconds=cycle, args=[file_name])
        self.sched.start()

    def save_db(self, file_name: str = 'data.cdb'):
        '''
            Data persistence, save the CyberDB database to the hard disk.

            parameter:

                file_name -- the name of the file generated by the backup, the file 
                suffix must be .cdb.
                
            Return Type: None
        '''
        prefix_name, suffix_name = file_name.rsplit('.', 1)

        if suffix_name != 'cdb':
            raise WrongFilenameCyberDBError(
                'Please enter a filename with a .cdb suffix.')

        # save to hard drive
        file_name_temp = prefix_name + '_temp.cdb'
        with open(file_name_temp, 'wb') as f:
            pickle.dump(self._data, f)
        shutil.move(file_name_temp, file_name)

    def load_db(self, file_name: str = 'data.cdb'):
        '''
            Load a file in .cdb format to load the CyberDB database backed up from 
            the hard disk back into memory.

            parameter:
                file_name -- the file name generated by data persistence, the file 
                suffix must be .cdb.
                
            Return Type: None
        '''
        prefix_name, suffix_name = file_name.rsplit('.', 1)

        if suffix_name != 'cdb':
            raise WrongFilenameCyberDBError('The file suffix must be cdb.')

        with open(file_name, 'rb') as f:
            self._data = pickle.load(f)

        # print('File {} loaded successfully.'.format(file_name))



import time
import inspect
import socket
from typing import List, Tuple, Dict

from obj_encrypt import Secret

from ..data import datas
from ..extensions import CyberDBError, WrongInputCyberDBError, WrongPasswordCyberDBError, WrongTableNameCyberDBError
from ..extensions.signature import Signature
from . import Stream


class Connection:
    '''
        TCP connection data type
    '''

    def __init__(
        self,
        s: socket.socket = None,
    ):
        self.s = s


class ConPool:
    '''
        Maintain connection pool.
    '''

    def __init__(self, host: str, port: str, dp: datas.DataParsing,
                 time_out: int = None):
        self._host = host
        self._port = port
        self._dp = dp
        self._time_out = time_out
        self._connections = []

    def get(self):
        '''
            Get the connection from the connection pool.
        '''
        while self._connections:
            # If the connection is full, the loop waits until a connection
            # is free.
            connection = self._connections.pop()
            time_now = int(time.time())
            if self._time_out:
                if time_now > connection['timestamp'] + self._time_out:
                    continue
            # Check if the server is down.
            r = confirm_the_connection(connection['s'], self._dp)
            if r['code'] == 1:
                return connection['s']
            else:
                pass

        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect((self._host, self._port))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s

    def put(self, s: socket.socket):
        '''
            Return the connection to the connection pool.
        '''
        # If a connection timeout is set, get the current timestamp.
        timestamp = None
        if self._time_out:
            timestamp = int(time.time())

        self._connections.insert(0, {
            's': s,
            'timestamp': timestamp
        })


def network(func):
    '''
        Network operations before and after encapsulating CyberDB data 
        structure methods.
    '''

    def wrapper(self, *args, **kw):
        stream = Stream(self._con.s, self._dp)
        client_obj = func(self, *args, **kw)
        stream.write(client_obj)

        server_obj = stream.read()
        if server_obj['code'] == 0:
            self._con.s.close()
            raise server_obj['Exception']

        if server_obj.get('content'):
            return server_obj['content']

    return wrapper


class CyberDict:
    '''
        A child object generated by a Proxy object for performing 
        Dictionaries operations. This object will perform remote 
        operations on the server-side CyberDB database. Shares the 
        same TCP connection with the Proxy object. The connection 
        will follow the connection of the Proxy. After the Proxy 
        releases the connection, the object will also lose the 
        connection. CyberDict can execute the get, setdefault, update, 
        keys, values, items, pop, popitem, clear methods and common 
        magic methods of Dictionaries, please refer to [Python 
        dictionary official documentation]
        (https://docs.python.org/3/library/stdtypes.html#mapping-types-dict).

        Iterate over CyberDict using a for loop with client space 
        complexity o(n), where n is the length of CyberDict.keys() .
    '''

    def __init__(
        self,
        table_name: str,
        dp: datas.DataParsing,
        con: Connection
    ):
        self._table_name = table_name
        self._dp = dp
        self._con = con
        self._route = '/cyberdict'

    @network
    def __repr__(self):
        return {
            'route': self._route + '/repr',
            'table_name': self._table_name
        }

    @network
    def __str__(self):
        return {
            'route': self._route + '/str',
            'table_name': self._table_name
        }

    @network
    def __len__(self):
        return {
            'route': self._route + '/len',
            'table_name': self._table_name
        }

    def __iter__(self):
        return self.generate()

    @network
    def __getitem__(self, key):
        return {
            'route': self._route + '/getitem',
            'table_name': self._table_name,
            'key': key
        }

    @network
    def __setitem__(self, key, value):
        return {
            'route': self._route + '/setitem',
            'table_name': self._table_name,
            'key': key,
            'value': value
        }

    @network
    def __delitem__(self, key):
        return {
            'route': self._route + '/delitem',
            'table_name': self._table_name,
            'key': key
        }

    @network
    def todict(self) -> Dict:
        '''
            Convert CyberDict to Dictionaries.
            
            Return Type: Dict
        '''
        return {
            'route': self._route + '/todict',
            'table_name': self._table_name
        }

    @network
    def get(self, key, default=None) -> any:
        return {
            'route': self._route + '/get',
            'table_name': self._table_name,
            'key': key,
            'default': default
        }

    @network
    def setdefault(self, key, default=None) -> None:
        return {
            'route': self._route + '/setdefault',
            'table_name': self._table_name,
            'key': key,
            'default': default
        }

    @network
    def update(self, dict2) -> None:
        return {
            'route': self._route + '/update',
            'table_name': self._table_name,
            'dict2': dict2,
        }

    @network
    def keys(self) -> List:
        return {
            'route': self._route + '/keys',
            'table_name': self._table_name
        }

    @network
    def values(self) -> List:
        return {
            'route': self._route + '/values',
            'table_name': self._table_name
        }

    @network
    def items(self) -> List[Tuple]:
        return {
            'route': self._route + '/items',
            'table_name': self._table_name
        }

    @network
    def pop(self, key, default=None) -> any:
        return {
            'route': self._route + '/pop',
            'table_name': self._table_name,
            'key': key,
            'default': default
        }

    @network
    def popitem(self) -> Tuple[any, any]:
        return {
            'route': self._route + '/popitem',
            'table_name': self._table_name
        }

    @network
    def clear(self) -> None:
        return {
            'route': self._route + '/clear',
            'table_name': self._table_name
        }

    def generate(self):
        for key in self.keys():
            yield key


class CyberList:
    '''
        A child object generated by a Proxy object for performing Lists 
        operations. This object will perform remote operations on the 
        server-side CyberDB database. Shares the same TCP connection 
        with the Proxy object. The connection will follow the connection 
        of the Proxy. After the Proxy releases the connection, the object 
        will also lose the connection. CyberList can execute the append, 
        extend, insert, pop, remove, count, index, reverse, sort, clear 
        methods and common magic methods of Lists, please refer to 
        [Python List Official Documentation]
        (https://docs.python.org/3/tutorial/datastructures.html#more-on-lists).

        The CyberList is iterated using a for loop, each iteration 
        will fetch the content from the server, and the space complexity 
        of the client is o(1).
    '''

    def __init__(
        self,
        table_name: str,
        dp: datas.DataParsing,
        con: Connection
    ):
        self._table_name = table_name
        self._dp = dp
        self._con = con
        self._route = '/cyberlist'

    @network
    def __repr__(self):
        return {
            'route': self._route + '/repr',
            'table_name': self._table_name
        }

    @network
    def __str__(self):
        return {
            'route': self._route + '/str',
            'table_name': self._table_name
        }

    @network
    def __len__(self):
        return {
            'route': self._route + '/len',
            'table_name': self._table_name
        }

    def __iter__(self):
        return self.generate()

    @network
    def __getitem__(self, index):
        return {
            'route': self._route + '/getitem',
            'table_name': self._table_name,
            'index': index
        }

    @network
    def __setitem__(self, index, value):
        return {
            'route': self._route + '/setitem',
            'table_name': self._table_name,
            'index': index,
            'value': value
        }

    @network
    def __delitem__(self, index):
        return {
            'route': self._route + '/delitem',
            'table_name': self._table_name,
            'index': index
        }

    @network
    def tolist(self) -> List:
        return {
            'route': self._route + '/tolist',
            'table_name': self._table_name
        }

    @network
    def append(self, value) -> None:
        return {
            'route': self._route + '/append',
            'table_name': self._table_name,
            'value': value
        }

    @network
    def extend(self, obj) -> None:
        if type(obj) == CyberList:
            obj = obj.tolist()

        return {
            'route': self._route + '/extend',
            'table_name': self._table_name,
            'obj': obj
        }

    @network
    def insert(self, index, value) -> None:
        return {
            'route': self._route + '/insert',
            'table_name': self._table_name,
            'index': index,
            'value': value
        }

    @network
    def pop(self, index: int = -1) -> any:
        return {
            'route': self._route + '/pop',
            'table_name': self._table_name,
            'index': index
        }

    @network
    def remove(self, value) -> None:
        return {
            'route': self._route + '/remove',
            'table_name': self._table_name,
            'value': value
        }

    @network
    def count(self, value) -> int:
        return {
            'route': self._route + '/count',
            'table_name': self._table_name,
            'value': value
        }

    @network
    def index(self, value) -> int:
        return {
            'route': self._route + '/index',
            'table_name': self._table_name,
            'value': value
        }

    @network
    def reverse(self) -> None:
        return {
            'route': self._route + '/reverse',
            'table_name': self._table_name
        }

    @network
    def sort(self, key=None, reverse=False) -> None:
        # Reference the lambda part to func.
        if key:
            if key.__code__.co_name == '<lambda>':
                key = 'func = ' + \
                    inspect.getsource(key).split('=')[1].rsplit(')')[0]
            # De-indent the code and reference the function to the func.
            elif key.__code__.co_name:
                code = inspect.getsource(key)
                num = 0
                for i in range(len(code)):
                    if code[i] != ' ':
                        num = i
                        break
                if i != 0:
                    code_new = ''
                    for line in code.splitlines():
                        code_new += line[num:] + '\n'
                else:
                    code_new = code
                key = code_new + '\nfunc = {}'.format(key.__code__.co_name)

        return {
            'route': self._route + '/sort',
            'table_name': self._table_name,
            'key': key,
            'reverse': reverse
        }

    @network
    def clear(self) -> None:
        return {
            'route': self._route + '/clear',
            'table_name': self._table_name,
        }

    def generate(self):
        for i in range(self.__len__()):
            yield self.__getitem__(i)


class Proxy:
    '''
        The Proxy object generated by the cyberdb.Client.get_proxy method 
        can operate on the CyberDB database and manage the TCP connections 
        of the CyberDict and CyberList sub-objects generated by the Proxy. 
        After the Proxy object is initialized, it can be used after executing 
        the Proxy.connect method. The Proxy object and its sub-objects will 
        perform remote operations on the server-side CyberDB database.
    '''

    def __init__(self, con_pool: ConPool, dp: datas.DataParsing):
        self._con_pool = con_pool
        self._dp = dp
        # The connection used by the proxy, the first is the reader and the
        # second is the writer.
        self._con = Connection()

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return True

    def connect(self):
        '''
            Get a TCP connection from the connection pool and bind it to the Proxy 
            object.
            
            Return Type: None
        '''
        # If the proxy already has a connection, the connection will be
        # returned to the connection pool first.
        if self._con.s != None:
            self._con_pool.put(self._con.s)

        s = self._con_pool.get()
        self._con.s = s

    def close(self):
        '''
            Cancel the TCP connection bound to the Proxy object and return it to 
            the connection pool. It needs to be reset before the next operation.
            Execute the Proxy.connect method.
            
            Return Type: None
        '''
        if self._con == None:
            raise CyberDBError(
                'The connection could not be closed, the proxy has not acquired a connection.')

        self._con_pool.put(self._con.s)
        self._con.s = None

    def create_cyberdict(self, table_name: str, content: dict = {}):
        '''
            Create a CyberDict table.
            
            Parameters:
            
                table_name – table name.
                
                content -- table content, needs to be a dictionary type, the 
                default is an empty dictionary.
                
            Return Type: None
        '''
        if type(table_name) != str:
            raise WrongInputCyberDBError('Please use str for the table name.')
        if type(content) != dict:
            raise WrongInputCyberDBError(
                'The input database table type is not a Python dictionary.')

        stream = Stream(self._con.s, self._dp)
        client_obj = {
            'route': '/create_cyberdict',
            'table_name': table_name,
            'content': content
        }
        stream.write(client_obj)

        server_obj = stream.read()

        if server_obj['code'] == 0:
            raise WrongTableNameCyberDBError(
                'Duplicate table names already exist!')

    def get_cyberdict(self, table_name: str) -> CyberDict:
        '''
            Get the CyberDict table.

            Parameters:

                table_name – table name.
                
            Return Type: CyberDict, which is a sub-object generated by Proxy, 
            which controls the TCP connection.
        '''
        if type(table_name) != str:
            raise WrongInputCyberDBError('Please use str for the table name.')

        stream = Stream(self._con.s, self._dp)

        client_obj = {
            'route': '/exam_cyberdict',
            'table_name': table_name
        }
        stream.write(client_obj)

        server_obj = stream.read()
        if server_obj['code'] == 0:
            raise WrongTableNameCyberDBError(
                '{} table does not exist.'.format(table_name))
        else:
            table = CyberDict(table_name, self._dp, self._con)
            return table

    def create_cyberlist(self, table_name: str, content: list = []):
        '''
            Create the CyberList table.

            Parameters:

                table_name – table name.

                content -- table content, it needs to be a list type, the default 
                is an empty list.
                
            Return Type: None
        '''
        if type(table_name) != str:
            raise WrongInputCyberDBError('Please use str for the table name.')
        if type(content) != type(list()):
            raise WrongInputCyberDBError(
                'The input database table type is not a Python dictionary.')

        stream = Stream(self._con.s, self._dp)
        client_obj = {
            'route': '/create_cyberlist',
            'table_name': table_name,
            'content': content
        }
        stream.write(client_obj)

        server_obj = stream.read()

        if server_obj['code'] == 0:
            raise WrongTableNameCyberDBError(
                'Duplicate table names already exist!')

    def get_cyberlist(self, table_name: str) -> CyberList:
        '''
            Get the CyberList table.

            Parameters:

                table_name – table name.
                
            Return type: CyberList, which is a sub-object generated by Proxy, which 
            controls the TCP connection.abase.
        '''
        if type(table_name) != str:
            raise WrongInputCyberDBError('Please use str for the table name.')

        stream = Stream(self._con.s, self._dp)

        client_obj = {
            'route': '/exam_cyberlist',
            'table_name': table_name
        }
        stream.write(client_obj)

        server_obj = stream.read()
        if server_obj['code'] == 0:
            raise WrongTableNameCyberDBError(
                '{} table does not exist.'.format(table_name))
        else:
            table = CyberList(table_name, self._dp, self._con)
            return table

    def print_tables(self):
        '''
            Print all tables in the CyberDB database.
            
            Return Type: None
        '''
        @network
        def get_tables(self):
            return {
                'route': '/print_tables'
            }

        r = get_tables(self)
        for line in r:
            print('table name: {}  type name: {}'.format(line[0], line[1]))

    def delete_table(self, table_name: str):
        '''
            Drop the table_name table in the CyberDB database.

            Parameters:

                table_name – the name of the table to drop.
                
            Return Type: None
        '''
        if type(table_name) != str:
            raise WrongInputCyberDBError('Please use str for the table name.')

        stream = Stream(self._con.s, self._dp)

        client_obj = {
            'route': '/exam_cyberlist',
            'table_name': table_name
        }
        stream.write(client_obj)

        server_obj = stream.read()
        if server_obj['code'] == 0:
            raise WrongTableNameCyberDBError(
                '{} table does not exist.'.format(table_name))

        client_obj = {
            'route': '/delete_table',
            'table_name': table_name
        }
        stream.write(client_obj)
        stream.read()


class Client:
    '''
        The cyberdb.Client object returned by the cyberdb.connect function 
        is used to generate the Proxy object.
    '''

    def __init__(self, con_pool: ConPool, dp: datas.DataParsing):
        self._con_pool = con_pool
        self._dp = dp

    def get_proxy(self) -> Proxy:
        '''
            Generate a Proxy object.

            Return Type: None
        '''
        proxy = Proxy(self._con_pool, self._dp)
        return proxy


def connect(host: str = '127.0.0.1', port: int = 9980, password:
            str = None, encrypt: bool = False, time_out: int = None) -> Client:
    '''
        Connect the client to the CyberDB server.

        Parameters:

            host -- the connection address, such as 127.0.0.1

            port -- connection port

            password -- connection password

            encrypt -- Whether to encrypt communication, if the server enables 
            encrypt to be True, it must be True here, and vice versa.

            time_out -- The timeout for each connection in the connection pool, 
            in seconds. Connections in the connection pool will be discarded 
            after time_out seconds of inactivity, and a new connection will be 
            generated next time. The connection pool will manage the 
            connections automatically, and the developer does not need to pay 
            attention to the details. If this parameter is None, there will be 
            no timeout, and the connection pool will maintain the connection 
            until it expires, after which a new connection will be regenerated.

        Return Type: Client
    '''
    if not password:
        raise WrongPasswordCyberDBError('The password cannot be empty.')
    if time_out and type(time_out) != int:
        raise CyberDBError('time_out must be an integer.')

    # Responsible for encrypting and decrypting objects.
    secret = Secret(key=password)
    # for digital signature
    signature = Signature(salt=password.encode())
    dp = datas.DataParsing(secret, signature, encrypt=encrypt)
    con_pool = ConPool(host, port, dp, time_out=time_out)

    # Synchronously test whether the connection is successful.
    s = con_pool.get()
    confirm_the_connection(s, dp)
    con_pool.put(s)

    client = Client(con_pool, dp)
    return client


def confirm_the_connection(s: socket.socket, dp: datas.DataParsing) -> dict:
    '''
        The connection is detected when the database connects for the first 
        time.
    '''
    try:
        stream = Stream(s, dp)

        client_obj = {
            'route': '/connect'
        }
        stream.write(client_obj)

        server_obj = stream.read()

        return {
            'code': 1,
            'content': server_obj
        }
    except Exception as e:
        return {
            'code': 0,
            'Exception': e
        }



'''
    Data used for TCP transmission
'''


import base64
import pickle

from obj_encrypt import Secret

from ..extensions.signature import Signature
from ..extensions import CyberDBError


def generate_client_obj():
    return {
        'route': None,
        'message': None
    }


def generate_server_obj():
    return {
        'code': None,
        'message': None
    }


errors_code = {
    2: 'Incorrect password or data tampering.'
}


class DataParsing:
    '''
        Convert TCP data and encrypted objects to each other.
    '''

    def __init__(self, secret: Secret, signature: Signature, encrypt: bool=False):
        self._secret = secret
        self._signature = signature
        self._encrypt = encrypt

    def data_to_obj(self, data):
        '''
            Restore TCP encrypted data as an object.
        '''
        # Determine whether to decrypt and verify the signature.
        if self._encrypt:
            try:
                data = base64.b64decode(data)
            except:
                return {
                    'code': 2,
                    'errors-code': errors_code[2]
                }

            try:
                data = self._secret.decrypt(data)
            except UnicodeDecodeError:
                return {
                    'code': 2,
                    'errors-code': errors_code[2]
                }

            # Check if the dictionary is intact.
            if type(data) != type(dict()):
                return {
                    'code': 2,
                    'errors-code': errors_code[2]
                }
            elif type(data.get('content')) != type(b'') or not data.get('header') or not \
                data.get('header').get('signature'):
                return {
                    'code': 2,
                    'errors-code': errors_code[2]
                }

            # Verify signature
            if self._signature.encrypt(data['content']) != data['header']['signature']:
                return {
                    'code': 2,
                    'errors-code': errors_code[2]
                }
                
            obj = pickle.loads(data['content'])
            return {
                'code': 1,
                'content': obj
            }
            
        else:
            try:
                data = base64.b64decode(data)
            except:
                return {
                    'code': 2,
                    'errors-code': errors_code[2]
                }
            
            try:
                obj = pickle.loads(data)
                return {
                    'code': 1,
                    'content': obj
                }
            except Exception as e:
                return {
                    'code': 2,
                    'errors-code': errors_code[2]
                }

    def obj_to_data(self, obj):
        '''
            Convert object to TCP transmission data.
        '''
        data = {
            'content': None,
            'header': {
                'signature': None
            }
        }
        
        # Determine whether to encrypt and sign.
        if self._encrypt:
            try:
                data['content'] = pickle.dumps(obj)
            except pickle.PickleError as e:
                raise CyberDBError('CyberDB does not support this data type.')
            
            data['header']['signature'] = self._signature.encrypt(data['content'])
            data = self._secret.encrypt(data)
            data = base64.b64encode(data)
            
        else:
            data = base64.b64encode(pickle.dumps(obj))

        return data





from threading import Thread


class MyThread(Thread):
    def __init__(self, target, args):
        super(MyThread, self).__init__()
        self.target = target
        self.args = args

    def run(self):
        self.result = self.target(*self.args)

    def get_result(self):
        try:
            return self.result
        except Exception:
            return None


class CyberDBError(RuntimeError):
    '''
        CyberDB basic error
    '''
    pass


class DisconCyberDBError(CyberDBError):
    '''
        TCP connection has been disconnected.
    '''
    pass


class WrongPasswordCyberDBError(CyberDBError):
    pass

class WrongFilenameCyberDBError(CyberDBError):
    pass

class BackupCyberDBError(CyberDBError):
    pass

class TypeCyberDBError(CyberDBError):
    pass

class WrongTableNameCyberDBError(CyberDBError):
    pass

class WrongInputCyberDBError(CyberDBError):
    pass



import asyncio

from obj_encrypt import Secret

from . import AioStream
from ..data import datas
from ..extensions import MyThread, CyberDBError
from ..extensions.signature import Signature


result = None # Check whether the connection is successful


class Connection:
    '''
        Asyncio TCP connection data type
    '''

    def __init__(
        self,
        reader: asyncio.streams.StreamReader=None,
        writer: asyncio.streams.StreamWriter=None
    ):
        self.reader = reader
        self.writer = writer


class ConPool:
    '''
        Maintain connection pool.
    '''
    
    def __init__(self, host: str, port: str, dp: datas.DataParsing):
        self._host = host
        self._port = port
        self._dp = dp
        self._connections = []

    async def get(self):
        '''
            Get the connection from the connection pool.
        '''
        if self._connections:
            # If the connection is full, the loop waits until a connection 
            # is free.
            while self._connections:
                reader, writer = self._connections.pop(0)
                if await self.exam(reader, writer):
                    return reader, writer

        reader, writer = await asyncio.open_connection(
        self._host, self._port)
        return reader, writer

    def put(self, reader: asyncio.streams.StreamReader, 
        writer: asyncio.streams.StreamWriter):
        '''
            Return the connection to the connection pool.
        '''
        self._connections.append((reader, writer))

    async def exam(self, reader: asyncio.streams.StreamReader, 
        writer: asyncio.streams.StreamWriter):
        '''
            Check if the connection is valid.
        '''
        # try:
        stream = AioStream(reader, writer, self._dp)
        client_obj = {
            'route': '/connect'
        }
        await stream.write(client_obj)

        server_obj = await stream.read()
        return True
        # except:
        #     return False


class CyberDict:

    def __init__(
        self, 
        table_name: str, 
        dp: datas.DataParsing,
        con: Connection
    ):
        self._table_name = table_name
        self._dp = dp
        self._con = con
        self._route = '/cyberdict'
        
    async def __getitem__(self, key):
        stream = AioStream(self._con.reader, self._con.writer, self._dp)

        client_obj = {
            'route': self._route + '/getitem',
            'table_name': self._table_name,
            'key': key
        }
        
        await stream.write(client_obj)

        server_obj = await stream.read()
        if server_obj['code'] == 0:
            self._con.writer.close()
            raise server_obj['Exception']

        return server_obj['content']


class CyberList:

    def __init__(
        self, 
        table_name: str, 
        dp: datas.DataParsing,
        con: Connection
    ):
        self._table_name = table_name
        self._dp = dp
        self._con = con
        self._route = '/cyberlist'


class Proxy:
    '''
        Database instance per session

        This object is not thread safe, please regenerate in each thread or 
        coroutine.
    '''

    def __init__(self, con_pool: ConPool, dp: datas.DataParsing):
        self._con_pool = con_pool
        self._dp = dp
        # The connection used by the proxy, the first is the reader and the 
        # second is the writer.
        self._con = Connection()

    async def connect(self):
        '''
            Get the latest connection from the connection pool.

            This method does not necessarily create a new connection, if the 
            connection of the connection pool is available, it will be 
            obtained directly.
        '''
        # If the proxy already has a connection, the connection will be 
        # returned to the connection pool first.
        if self._con.reader != None and self._con.writer != None:
            self._con_pool.put(self._con.reader, self._con.writer)
            
        reader, writer = await self._con_pool.get()
        self._con.reader = reader
        self._con.writer = writer

    async def close(self):
        '''
            Return the connection to the connection pool.
        '''
        if self._con.reader == None or self._con == None:
            raise CyberDBError('The connection could not be closed, the proxy has not acquired a connection.')

        self._con_pool.put(self._con.reader, self._con.writer)
        self._con.reader = None
        self._con.writer = None

    async def create_cyberdict(self, table_name: str, content: dict={}):
        if type(table_name) != type(''):
            raise CyberDBError('Please use str for the table name.')
        if type(content) != type(dict()):
            raise CyberDBError('The input database table type is not a Python dictionary.')

        stream = AioStream(self._con.reader, self._con.writer, self._dp)
        client_obj = {
            'route': '/create_cyberdict',
            'table_name': table_name,
            'content': content
        }
        await stream.write(client_obj)

        server_obj = await stream.read()

        if server_obj['code'] == 0:
            raise CyberDBError('Duplicate table names already exist!')

        return server_obj

    async def get_cyberdict(self, table_name: str) -> CyberDict:
        '''
            Get the network object of the table from the database.
        '''
        if type(table_name) != type(''):
            raise CyberDBError('Please use str for the table name.')
        
        stream = AioStream(self._con.reader, self._con.writer, self._dp)

        client_obj = {
            'route': '/exam_cyberdict',
            'table_name': table_name
        }
        await stream.write(client_obj)

        server_obj = await stream.read()
        if server_obj['code'] == 0:
            raise CyberDBError('{} table does not exist.'.format(table_name))
        else:
            table = CyberDict(table_name, self._dp, self._con)
            return table


class Client:
    '''
        This object blocks execution and is recommended to be used as a 
        global variable instead of running in an async function.
    '''
    
    def __init__(self, con_pool: ConPool, dp: datas.DataParsing):
        self._con_pool = con_pool
        self._dp = dp

    def get_proxy(self) -> Proxy:
        '''
            Get proxy data.
        '''
        proxy = Proxy(self._con_pool, self._dp)
        return proxy

    def check_connection_pool(self):
        pass


def connect(host: str='127.0.0.1', port: int=9980, password: 
    str=None) -> Client:
    '''
        Connect to the CyberDB server via TCP, this method will run 
        synchronously.
    '''
    if not password:
        raise RuntimeError('The password cannot be empty.')

    # Responsible for encrypting and decrypting objects.
    secret = Secret(key=password)
    # for digital signature
    signature = Signature(salt=password.encode())
    dp = datas.DataParsing(secret, signature)
    con_pool = ConPool(host, port, dp)

    # Synchronously test whether the connection is successful.
    t = MyThread(target=asyncio.run, 
        args=(confirm_the_connection(con_pool, dp), ) )
    t.daemon = True
    t.start()
    t.join()
    result = t.get_result()
    if result['code'] == 0:
        raise result['Exception']

    client = Client(con_pool, dp)
    return client


async def confirm_the_connection(con_pool: ConPool, dp: datas.DataParsing) -> \
    dict:
    '''
        The connection is detected when the database connects for the first 
        time.
    '''
    try:
        reader, writer = await con_pool.get()
        stream = AioStream(reader, writer, dp)

        client_obj = {
            'route': '/connect'
        }
        await stream.write(client_obj)

        server_obj = await stream.read()
        
        writer.close()
        
        return {
            'code': 1,
            'content': server_obj
        }
    except (ConnectionRefusedError, CyberDBError) as e:
        return {
            'code': 0,
            'Exception': e
        }
    # except Exception as e:
    #     return {
    #         'code': 0,
    #         'Exception': CyberDBError('Incorrect address, port or password for database.')
    #     }





import random


seed = '1234567890abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'


def generate(num: int):
    '''
        Generate num random strings.
    '''
    text = ''
    for i in range(num):
        text += random.choice(seed)
    return text



import inspect
import datetime

from . import AioStream
from ..data import datas
from ..extensions import nonce, WrongPasswordCyberDBError


MAP = {}


def bind(path):
    def decorator(func):
        async def wrapper(self, *args, **kw):
            server_obj = await func(self, *args, **kw)
            server_obj['password'] = self._dp._secret.key
            
            await self._stream.write(server_obj)

        MAP[path] = wrapper
        return wrapper
    return decorator


class Route:
    '''
        TCP event mapping.
    '''

    def __init__(self, db: dict, dp: datas.DataParsing, stream: AioStream,
                 print_log: bool = False):
        self._db = db
        self._dp = dp
        self._stream = stream
        self._print_log = print_log

    async def find(self):
        '''
            Loop accepting client requests.
        '''
        addr = self._stream.get_addr()
        addr = '{}:{}'.format(addr[0], addr[1])

        while True:
            client_obj = await self._stream.read()
            self._client_obj = client_obj

            # Check if the password is correct.
            if self._stream._dp._secret.key != client_obj['password']:
                self._stream._writer.close()
                raise WrongPasswordCyberDBError(
                    'The password entered by the client is incorrect.')

            if self._print_log:
                print('{}  {}  {}'.format(
                    datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'),
                    addr, client_obj['route']))

            # Jump to the specified function by routing.
            func = MAP[client_obj['route']]
            # Go to the corresponding routing function.
            await func(self)

    @bind('/connect')
    async def connect(self):
        server_obj = {
            'code': 1,
            'message': 'connection succeeded.'
        }
        return server_obj

    @bind('/create_cyberdict')
    async def create_cyberdict(self):
        table_name = self._client_obj['table_name']
        content = self._client_obj['content']
        if self._db.get(table_name) == None:
            # New CyberDict
            self._db[table_name] = content
            server_obj = {
                'code': 1
            }
        else:
            server_obj = {
                'code': 0
            }

        return server_obj

    @bind('/create_cyberlist')
    async def create_cyberlist(self):
        table_name = self._client_obj['table_name']
        content = self._client_obj['content']
        if self._db.get(table_name) == None:
            # New CyberList
            self._db[table_name] = content
            server_obj = {
                'code': 1
            }
        else:
            server_obj = {
                'code': 0
            }

        return server_obj

    @bind('/exam_cyberdict')
    async def exam_cyberdict(self):
        '''
            Check if the table in the database exists.
        '''

        table_name = self._client_obj['table_name']
        if self._db.get(table_name) != None:
            server_obj = {
                'code': 1
            }
        else:
            server_obj = {
                'code': 0
            }

        return server_obj

    @bind('/exam_cyberlist')
    async def exam_cyberlist(self):
        '''
            Check if the table in the database exists.
        '''

        table_name = self._client_obj['table_name']
        if self._db.get(table_name) != None:
            server_obj = {
                'code': 1
            }
        else:
            server_obj = {
                'code': 0
            }

        return server_obj

    @bind('/print_tables')
    async def print_tables(self):
        inform = []
        for table_name in self._db:
            if type(self._db[table_name]) == dict:
                type_name = 'CyberDict'
            elif type(self._db[table_name]) == list:
                type_name = 'CyberList'
            inform.append((table_name, type_name))
            
        server_obj = {
            'code': 1,
            'content': inform
        }
        
        return server_obj

    @bind('/delete_table')
    async def delete_table(self):
        table_name = self._client_obj['table_name']
        try:
            del self._db[table_name]
            server_obj = {
                'code': 1
            }
        except Exception as e:
            server_obj = {
                'code': 0,
                'Exception': e
            }
        
        return server_obj

    @bind('/cyberdict/repr')
    async def dict_repr(self):
        table_name = self._client_obj['table_name']
        try:
            r = repr(self._db[table_name])
            server_obj = {
                'code': 1,
                'content': r
            }
        except Exception as e:
            server_obj = {
                'code': 0,
                'Exception': e
            }

        return server_obj

    @bind('/cyberdict/str')
    async def dict_str(self):
        table_name = self._client_obj['table_name']
        try:
            r = str(self._db[table_name])
            server_obj = {
                'code': 1,
                'content': r
            }
        except Exception as e:
            server_obj = {
                'code': 0,
                'Exception': e
            }

        return server_obj

    @bind('/cyberdict/len')
    async def dict_len(self):
        table_name = self._client_obj['table_name']
        try:
            r = len(self._db[table_name])
            server_obj = {
                'code': 1,
                'content': r
            }
        except Exception as e:
            server_obj = {
                'code': 0,
                'Exception': e
            }

        return server_obj

    @bind('/cyberdict/getitem')
    async def dict_getitem(self):
        table_name = self._client_obj['table_name']
        key = self._client_obj['key']
        try:
            r = self._db[table_name][key]
            server_obj = {
                'code': 1,
                'content': r
            }
        except Exception as e:
            server_obj = {
                'code': 0,
                'Exception': e
            }

        return server_obj

    @bind('/cyberdict/setitem')
    async def dict_setitem(self):
        table_name = self._client_obj['table_name']
        key = self._client_obj['key']
        value = self._client_obj['value']
        try:
            self._db[table_name][key] = value
            server_obj = {
                'code': 1,
            }
        except Exception as e:
            server_obj = {
                'code': 0,
                'Exception': e
            }

        return server_obj

    @bind('/cyberdict/delitem')
    async def dict_delitem(self):
        table_name = self._client_obj['table_name']
        key = self._client_obj['key']
        try:
            del self._db[table_name][key]
            server_obj = {
                'code': 1,
            }
        except Exception as e:
            server_obj = {
                'code': 0,
                'Exception': e
            }

        return server_obj

    @bind('/cyberdict/todict')
    async def todict(self):
        table_name = self._client_obj['table_name']
        try:
            r = self._db[table_name]
            server_obj = {
                'code': 1,
                'content': r
            }
        except Exception as e:
            server_obj = {
                'code': 0,
                'Exception': e
            }

        return server_obj

    @bind('/cyberdict/get')
    async def dict_get(self):
        table_name = self._client_obj['table_name']
        key = self._client_obj['key']
        default = self._client_obj['default']
        try:
            r = self._db[table_name].get(key, default)
            server_obj = {
                'code': 1,
                'content': r
            }
        except Exception as e:
            server_obj = {
                'code': 0,
                'Exception': e
            }

        return server_obj

    @bind('/cyberdict/setdefault')
    async def dict_setdefault(self):
        table_name = self._client_obj['table_name']
        key = self._client_obj['key']
        default = self._client_obj['default']
        try:
            r = self._db[table_name].setdefault(key, default)
            server_obj = {
                'code': 1,
                'content': r
            }
        except Exception as e:
            server_obj = {
                'code': 0,
                'Exception': e
            }

        return server_obj

    @bind('/cyberdict/update')
    async def dict_update(self):
        table_name = self._client_obj['table_name']
        dict2 = self._client_obj['dict2']
        try:
            self._db[table_name].update(dict2)
            server_obj = {
                'code': 1
            }
        except Exception as e:
            server_obj = {
                'code': 0,
                'Exception': e
            }

        return server_obj

    @bind('/cyberdict/keys')
    async def dict_keys(self):
        table_name = self._client_obj['table_name']
        try:
            r = list(self._db[table_name].keys())
            server_obj = {
                'code': 1,
                'content': r
            }
        except Exception as e:
            server_obj = {
                'code': 0,
                'Exception': e
            }

        return server_obj

    @bind('/cyberdict/values')
    async def dict_values(self):
        table_name = self._client_obj['table_name']
        try:
            r = list(self._db[table_name].values())
            server_obj = {
                'code': 1,
                'content': r
            }
        except Exception as e:
            server_obj = {
                'code': 0,
                'Exception': e
            }

        return server_obj

    @bind('/cyberdict/items')
    async def dict_items(self):
        table_name = self._client_obj['table_name']
        try:
            r = list(self._db[table_name].items())
            server_obj = {
                'code': 1,
                'content': r
            }
        except Exception as e:
            server_obj = {
                'code': 0,
                'Exception': e
            }

        return server_obj

    @bind('/cyberdict/pop')
    async def dict_pop(self):
        table_name = self._client_obj['table_name']
        key = self._client_obj['key']
        default = self._client_obj['default']

        try:
            r = self._db[table_name].pop(key, default)
            server_obj = {
                'code': 1,
                'content': r
            }
        except Exception as e:
            server_obj = {
                'code': 0,
                'Exception': e
            }

        return server_obj

    @bind('/cyberdict/popitem')
    async def dict_popitem(self):
        table_name = self._client_obj['table_name']
        try:
            r = self._db[table_name].popitem()
            server_obj = {
                'code': 1,
                'content': r
            }
        except Exception as e:
            server_obj = {
                'code': 0,
                'Exception': e
            }

        return server_obj

    @bind('/cyberdict/clear')
    async def dict_clear(self):
        table_name = self._client_obj['table_name']
        try:
            self._db[table_name].clear()
            server_obj = {
                'code': 1
            }
        except Exception as e:
            server_obj = {
                'code': 0,
                'Exception': e
            }

        return server_obj

    @bind('/cyberlist/repr')
    async def list_repr(self):
        table_name = self._client_obj['table_name']
        try:
            r = repr(self._db[table_name])
            server_obj = {
                'code': 1,
                'content': r
            }
        except Exception as e:
            server_obj = {
                'code': 0,
                'Exception': e
            }

        return server_obj

    @bind('/cyberlist/str')
    async def list_str(self):
        table_name = self._client_obj['table_name']
        try:
            r = str(self._db[table_name])
            server_obj = {
                'code': 1,
                'content': r
            }
        except Exception as e:
            server_obj = {
                'code': 0,
                'Exception': e
            }

        return server_obj

    @bind('/cyberlist/len')
    async def list_len(self):
        table_name = self._client_obj['table_name']
        try:
            r = len(self._db[table_name])
            server_obj = {
                'code': 1,
                'content': r
            }
        except Exception as e:
            server_obj = {
                'code': 0,
                'Exception': e
            }

        return server_obj

    @bind('/cyberlist/getitem')
    async def list_getitem(self):
        table_name = self._client_obj['table_name']
        index = self._client_obj['index']
        try:
            r = self._db[table_name][index]
            server_obj = {
                'code': 1,
                'content': r
            }
        except Exception as e:
            server_obj = {
                'code': 0,
                'Exception': e
            }

        return server_obj

    @bind('/cyberlist/setitem')
    async def list_setitem(self):
        table_name = self._client_obj['table_name']
        index = self._client_obj['index']
        value = self._client_obj['value']
        try:
            self._db[table_name][index] = value
            server_obj = {
                'code': 1,
            }
        except Exception as e:
            server_obj = {
                'code': 0,
                'Exception': e
            }

        return server_obj

    @bind('/cyberlist/delitem')
    async def list_delitem(self):
        table_name = self._client_obj['table_name']
        index = self._client_obj['index']
        try:
            del self._db[table_name][index]
            server_obj = {
                'code': 1,
            }
        except Exception as e:
            server_obj = {
                'code': 0,
                'Exception': e
            }

        return server_obj

    @bind('/cyberlist/tolist')
    async def tolist(self):
        table_name = self._client_obj['table_name']
        try:
            r = self._db[table_name]
            server_obj = {
                'code': 1,
                'content': r
            }
        except Exception as e:
            server_obj = {
                'code': 0,
                'Exception': e
            }

        return server_obj

    @bind('/cyberlist/append')
    async def list_append(self):
        table_name = self._client_obj['table_name']
        value = self._client_obj['value']
        try:
            self._db[table_name].append(value)
            server_obj = {
                'code': 1
            }
        except Exception as e:
            server_obj = {
                'code': 0,
                'Exception': e
            }

        return server_obj

    @bind('/cyberlist/extend')
    async def list_extend(self):
        table_name = self._client_obj['table_name']
        obj = self._client_obj['obj']
        try:
            self._db[table_name].extend(obj)
            server_obj = {
                'code': 1
            }
        except Exception as e:
            server_obj = {
                'code': 0,
                'Exception': e
            }

        return server_obj

    @bind('/cyberlist/insert')
    async def list_insert(self):
        table_name = self._client_obj['table_name']
        index = self._client_obj['index']
        value = self._client_obj['value']
        try:
            self._db[table_name].insert(index, value)
            server_obj = {
                'code': 1
            }
        except Exception as e:
            server_obj = {
                'code': 0,
                'Exception': e
            }

        return server_obj

    @bind('/cyberlist/pop')
    async def list_pop(self):
        table_name = self._client_obj['table_name']
        index = self._client_obj['index']
        try:
            self._db[table_name].pop(index)
            server_obj = {
                'code': 1
            }
        except Exception as e:
            server_obj = {
                'code': 0,
                'Exception': e
            }

        return server_obj

    @bind('/cyberlist/remove')
    async def list_remove(self):
        table_name = self._client_obj['table_name']
        value = self._client_obj['value']
        try:
            self._db[table_name].remove(value)
            server_obj = {
                'code': 1
            }
        except Exception as e:
            server_obj = {
                'code': 0,
                'Exception': e
            }

        return server_obj

    @bind('/cyberlist/count')
    async def list_count(self):
        table_name = self._client_obj['table_name']
        value = self._client_obj['value']
        try:
            r = self._db[table_name].count(value)
            server_obj = {
                'code': 1,
                'content': r
            }
        except Exception as e:
            server_obj = {
                'code': 0,
                'Exception': e
            }

        return server_obj

    @bind('/cyberlist/index')
    async def list_index(self):
        table_name = self._client_obj['table_name']
        value = self._client_obj['value']
        try:
            r = self._db[table_name].index(value)
            server_obj = {
                'code': 1,
                'content': r
            }
        except Exception as e:
            server_obj = {
                'code': 0,
                'Exception': e
            }

        return server_obj

    @bind('/cyberlist/reverse')
    async def list_reverse(self):
        table_name = self._client_obj['table_name']
        try:
            self._db[table_name].reverse()
            server_obj = {
                'code': 1
            }
        except Exception as e:
            server_obj = {
                'code': 0,
                'Exception': e
            }

        return server_obj

    @bind('/cyberlist/sort')
    async def list_sort(self):
        table_name = self._client_obj['table_name']
        key = self._client_obj['key']
        
        if key:
            # Dynamic get function.
            loc = locals()
            key = exec('{}'.format(key))
            key = loc['func']
            
        reverse = self._client_obj['reverse']
        try:
            self._db[table_name].sort(key=key, reverse=reverse)
            server_obj = {
                'code': 1
            }
        except Exception as e:
            server_obj = {
                'code': 0,
                'Exception': e
            }

        return server_obj

    @bind('/cyberlist/clear')
    async def list_clear(self):
        table_name = self._client_obj['table_name']
        try:
            self._db[table_name].clear()
            server_obj = {
                'code': 1
            }
        except Exception as e:
            server_obj = {
                'code': 0,
                'Exception': e
            }

        return server_obj



import setuptools


with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="CyberDB",
    version="0.4.3",
    author="Cyberbolt",
    author_email="dtconlyone@gmail.com",
    description="CyberDB is a lightweight Python in-memory database. It is designed to use Python's built-in data structures Dictionaries, Lists for data storage, efficient communication through Socket TCP, and provide data persistence. This module is often used in Gunicorn inter-process communication, distributed computing and other fields.",
    long_description=long_description,
    long_description_content_type="text/markdown",    
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'APScheduler>=3.9.1',
        'obj-encrypt==0.3.0'
    ]    
)


import math
import socket
import asyncio

from ..data import datas
from ..extensions import CyberDBError, DisconCyberDBError, \
    WrongPasswordCyberDBError


SLICE_SIZE = 4096


class AioStream:
    '''
        Encapsulates TCP read and write and object encryption.
    '''

    def __init__(
        self,
        reader: asyncio.streams.StreamReader,
        writer: asyncio.streams.StreamWriter,
        dp: datas.DataParsing
    ):
        self._reader = reader
        self._writer = writer
        self._dp = dp

    async def read(self) -> dict:
        reader, writer = self._reader, self._writer

        # Gets the number of times TCP loops to receive data.
        number_of_times = await reader.read(SLICE_SIZE)

        # The client actively disconnects.
        if number_of_times == b'':
            raise DisconCyberDBError('The TCP connection was disconnected by the other end.')

        r = self._dp.data_to_obj(number_of_times)
        if r['code'] != 1:
            writer.close()
            raise WrongPasswordCyberDBError(r['errors-code'])
        number_of_times = r['content']

        # Informs the client that it is ready to receive data.
        ready = self._dp.obj_to_data('ready')
        writer.write(ready)
        await writer.drain()

        buffer = [await reader.read(SLICE_SIZE) for i in range(number_of_times)]
        data = b''.join(buffer) # Splice into complete data.

        r = self._dp.data_to_obj(data)
        if r['code'] != 1:
            writer.close()
            raise WrongPasswordCyberDBError(r['errors-code'])

        return r['content']

    async def write(self, obj: dict):
        reader, writer = self._reader, self._writer

        data = self._dp.obj_to_data(obj)

        # The number of times the other end TCP loops to 
        # receive data.
        number_of_times = self._dp.obj_to_data(
            math.ceil(len(data) / SLICE_SIZE)
        )
        writer.write(number_of_times)
        await writer.drain()

        # Get the client's readiness status.
        ready = await reader.read(SLICE_SIZE)
        r = self._dp.data_to_obj(ready)
        if r['code'] != 1:
            writer.close()
            raise WrongPasswordCyberDBError(r['errors-code'])
        ready = r['content']

        if ready != 'ready':
            raise RuntimeError('error')

        writer.write(data)
        await writer.drain()

    def get_addr(self):
        return self._writer.get_extra_info('peername')


class Stream:
    '''
        Synchronous TCP read and write
    '''

    def __init__(self, s: socket.socket, dp: datas.DataParsing):
        self._s = s
        self._dp = dp
        
    def read(self) -> dict:
        # Gets the number of times TCP loops to receive data.
        number_of_times = self._s.recv(SLICE_SIZE)

        # The client actively disconnects.
        if number_of_times == b'':
            raise DisconCyberDBError('The TCP connection was disconnected by the other end.')

        r = self._dp.data_to_obj(number_of_times)
        if r['code'] != 1:
            self._s.close()
            raise WrongPasswordCyberDBError(r['errors-code'])
        number_of_times = r['content']

        # Informs the client that it is ready to receive data.
        ready = self._dp.obj_to_data('ready')
        self._s.sendall(ready)

        buffer = [self._s.recv(SLICE_SIZE) for i in range(number_of_times)]
        data = b''.join(buffer) # Splice into complete data.

        r = self._dp.data_to_obj(data)
        if r['code'] != 1:
            self._s.close()
            raise WrongPasswordCyberDBError(r['errors-code'])

        return r['content']
    
    def write(self, obj: dict):
        obj['password'] = self._dp._secret.key
        
        data = self._dp.obj_to_data(obj)
        
        # The number of times the other end TCP loops to 
        # receive data.
        number_of_times = self._dp.obj_to_data(
            math.ceil(len(data) / SLICE_SIZE)
        )
        try:
            self._s.sendall(number_of_times)
        except BrokenPipeError as e:
            raise DisconCyberDBError('The TCP connection has been lost, please run proxy.connect() to regain the connection.')
        
        # Get the client's readiness status.
        ready = self._s.recv(SLICE_SIZE)
        try:
            r = self._dp.data_to_obj(ready)
        except EOFError as e:
            raise DisconCyberDBError('The TCP connection has been lost, please run proxy.connect() to regain the connection.')
        
        if r['code'] != 1:
            self._s.close()
            raise WrongPasswordCyberDBError(r['errors-code'])
        ready = r['content']

        if ready != 'ready':
            raise RuntimeError('error')
        
        self._s.sendall(data)
    


import hashlib


class Signature:
    '''
        This class is based on the SHA-256 algorithm for signing and checking data integrity.
    '''

    def __init__(self, salt: bytes, iterations: int=1):
        '''
            salt -- SHA-256 salt\n
            iterations -- The number of iterations of the SHA-256 algorithm.
        '''
        if type(salt) != type(b'1'):
            raise RuntimeError('Salt must be bytes.')
        self._salt = salt
        self._iterations = iterations

    def encrypt(self, content: bytes):
        '''
            SHA-256 encryption
        '''
        if type(content) != type(b'1'):
            raise RuntimeError('Content must be bytes.')
        dk = hashlib.pbkdf2_hmac('sha256', content, self._salt, self._iterations)
        return dk.hex()